/* 
**  mod_vhost_mysql.c -- Apache vhost_mysql2 module
**
**  (Badly) adapted to work with apache2 by Stefan de Konink <stefandekonink@xs4all.nl>
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"
#include "mysql.h"

#define VHOST_MYSQL_VERSION "0.10"
module AP_MODULE_DECLARE_DATA vhost_mysql_module;

MYSQL mysql, *mysql_p = NULL;

typedef struct {
    int enable;
    const char *db;
    const char *db_host;
    unsigned int db_port;
    const char *db_user;
    const char *db_pass;
    const char *db_socket;
    const char *table_name;
    const char *host_col;
    const char *docroot_col;
} vh_mysql_config;

static void *vh_mysql_create_config(apr_pool_t *p, server_rec *s)
{
	vh_mysql_config *conf = 
	    (vh_mysql_config *) apr_pcalloc(p, sizeof(vh_mysql_config));
	conf->enable = 0;
	conf->db = "vh";
	conf->db_host = "localhost";
	conf->db_port = 3306;
	conf->db_user = "root";
	conf->db_pass = "";
	conf->db_socket = NULL;
	conf->table_name = "vh";
	conf->host_col = "servername";
	conf->docroot_col = "docroot";
	return (void *) conf;
}

/*static void vh_mysql_initialize(server_rec *s, apr_pool_t *p) {
    apr_add_version_component("mod_vhost_mysql/" VHOST_MYSQL_VERSION);
}*/

static int open_link(vh_mysql_config *conf) {
    mysql_p = mysql_real_connect(&mysql, conf->db_host, conf->db_user, conf->db_pass, conf->db, conf->db_port, conf->db_socket, 0);
    if(!mysql_p)
	return HTTP_INTERNAL_SERVER_ERROR;
    return OK;
}

static apr_status_t vh_mysql_child_exit(void *data) {
    server_rec *s = data;
    vh_mysql_config *conf = 
	(vh_mysql_config *)ap_get_module_config(s->module_config, &vhost_mysql_module);
    if(conf->enable && (mysql_p != NULL))
	mysql_close(mysql_p);
}

static void vh_mysql_child_init(apr_pool_t *p, server_rec *s) {
vh_mysql_config *conf = 
      (vh_mysql_config *)ap_get_module_config(s->module_config, &vhost_mysql_module);
    int rc;
    if(!conf->enable)
	return;
    mysql_init(&mysql);
    if(open_link(conf) != OK) {
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, s, 
		     "mod_vhost_mysql MySQL connection failed.");
    }
    apr_pool_cleanup_register(p, s, vh_mysql_child_exit, vh_mysql_child_exit);
}

static apr_status_t cleanup_mysql_result(void *result){
    if(result != NULL)
	mysql_free_result((MYSQL_RES *)result);
}

static void register_cleanup_mysql_result (apr_pool_t *p, MYSQL_RES *result){
    apr_pool_cleanup_register(p, (void *)result, cleanup_mysql_result, cleanup_mysql_result);
}

static char *quote_sql(apr_pool_t *p, const char *str) {
  char *new;
  int  i, j, n;
  for (i=0, n=0; str[i]; i++) {
    if (str[i] == '\'') n++;
  }
  n += i + 1;
  new = apr_pcalloc(p, n + 2);

  new[0] = '\'';
  for (i=0, j=1; str[i]; i++) {
    new[j++] = str[i];
    if (str[i] == '\'') new[j++] = str[i];
  }
  new[i+1] = '\'';
  new[i+2] = 0;
  return new;
}

static char *construct_sql(apr_pool_t *p, vh_mysql_config *conf, const char *hostname){
    char *sql;
    sql = apr_psprintf(p, "SELECT %s FROM %s WHERE %s = %s",
		      conf->docroot_col, conf->table_name,
		      conf->host_col,
		      quote_sql(p, hostname));
    return sql;
}

static int vh_mysql_trans_handler(request_rec *r) {
    vh_mysql_config *conf = 
      (vh_mysql_config *)ap_get_module_config(r->server->module_config, &vhost_mysql_module);
    MYSQL_RES *result;
    MYSQL_ROW row;
    char *sql;
    char *localpath;
    if(!conf->enable)
	return DECLINED;
    if(mysql_ping(&mysql) != 0){
	if(open_link(conf) == OK) {
	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, 0, r, "reconnect to MySQL success");
	}
	else {
	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, 0, r, "reconnect to MySQL fail");
	    return HTTP_INTERNAL_SERVER_ERROR;
	}
    }
    sql = construct_sql(r->pool,conf, (const char *)ap_get_server_name(r));
//Rewrite...
//	sql = construct_sql(r->pool,conf, (const char *)r->server->server_hostname);
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r, "SQL: %s", sql);
    if(mysql_query(mysql_p, sql) != 0){
	ap_log_rerror(APLOG_MARK,  APLOG_NOERRNO|APLOG_ERR, 0, r, "mod_vhost_mysql SELECT query failed: %s", sql);
	return HTTP_INTERNAL_SERVER_ERROR;
    }
    result = mysql_store_result(mysql_p);
    register_cleanup_mysql_result(r->pool, result);
    row = mysql_fetch_row(result);
    if(row != NULL){
	localpath = apr_pstrcat(r->pool, row[0], r->uri, NULL);
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r, "http://%s%s -> %s", ap_get_server_name(r), r->uri, localpath);
	r->filename = localpath;
	return OK;
    }
    return DECLINED;
}

static const char *set_engine(cmd_parms *parms, void *dummy, int flag){
    vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
    conf->enable = flag;
    return NULL;
}

static const char *set_db_name(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->db = arg;
//	conf->db = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const char *set_db_host(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->db_host = arg;
//	conf->db_host = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const char *set_db_port(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->db_port = (unsigned int)atoi(arg);
	return NULL;
}

static const char *set_db_user(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->db_user = arg;
//	conf->db_user = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const char *set_db_pass(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->db_pass = arg;
//	conf->db_pass = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const char *set_db_socket(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->db_socket = arg;
//	conf->db_socket = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const char *set_table_name(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->table_name = arg;
//	conf->table_name = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const char *set_host_col(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->host_col = arg;
//	conf->host_col = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const char *set_docroot_col(cmd_parms *parms, void *dummy, const char *arg)
{
	vh_mysql_config *conf = 
	  (vh_mysql_config *)ap_get_module_config(parms->server->module_config, &vhost_mysql_module);
	conf->docroot_col = arg;
//	conf->docroot_col = apr_pstrdup(parms->pool, arg);
	return NULL;
}

static const command_rec vh_mysql_cmds[] = {
	AP_INIT_FLAG("VhMySQLEngine", set_engine, NULL, RSRC_CONF, "On or Off mod_vhost_mysql engine."),
	AP_INIT_TAKE1("VhMySQLTableName", set_table_name, NULL, RSRC_CONF, "set table name"),
	AP_INIT_TAKE1("VhMySQLDBName", set_db_name, NULL, RSRC_CONF, "set database name"),
	AP_INIT_TAKE1("VhMySQLDBHost", set_db_host, NULL, RSRC_CONF, "set database hostname"),
	AP_INIT_TAKE1("VhMySQLDBPort", set_db_port, NULL, RSRC_CONF, "set database port number"),
	AP_INIT_TAKE1("VhMySQLDBUser", set_db_user, NULL, RSRC_CONF, "set database user"),
	AP_INIT_TAKE1("VhMySQLDBPassword", set_db_pass, NULL, RSRC_CONF, "set database password"),
	AP_INIT_TAKE1("VhMySQLDBSocket", set_db_socket, NULL, RSRC_CONF, "set database socket file"),
	AP_INIT_TAKE1("VhMySQLServerNameField", set_host_col, NULL, RSRC_CONF, "set servername column"),
	AP_INIT_TAKE1("VhMySQLDocRootField", set_docroot_col, NULL, RSRC_CONF, "set docroot column"),
    {NULL}
};

static void vh_register_hooks(apr_pool_t *p) {
	ap_hook_child_init(vh_mysql_child_init, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_translate_name(vh_mysql_trans_handler, NULL, NULL, APR_HOOK_MIDDLE);
}



/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA  vhost_mysql_module = {
    STANDARD20_MODULE_STUFF, 
//    vh_mysql_initialize,   /* module initializer                  */
    NULL,                /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    vh_mysql_create_config,    /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    vh_mysql_cmds,         /* table of config file commands       */
    vh_register_hooks		/* set up other request processing hooks */
   
//    NULL,                  /* [#8] MIME-typed-dispatched handlers */
//    vh_mysql_trans_handler, /* [#1] URI to filename translation    */
//    NULL,                  /* [#4] validate user id from request  */
//    NULL,                  /* [#5] check if the user is ok _here_ */
//    NULL,                  /* [#3] check access by host address   */
//    NULL,                  /* [#6] determine MIME type            */
//    NULL,                  /* [#7] pre-run fixups                 */
//    NULL,                  /* [#9] log a transaction              */
//    NULL,                  /* [#2] header parser                  */
//    vh_mysql_child_init,   /* child_init                          */
//    vh_mysql_child_exit,   /* child_exit                          */
//    NULL                   /* [#0] post read-request              */
//#ifdef EAPI
//   ,NULL,                  /* EAPI: add_module                    */
//    NULL,                  /* EAPI: remove_module                 */
//    NULL,                  /* EAPI: rewrite_command               */
//    NULL                   /* EAPI: new_connection                */
//#endif
};

